# vim: ft=python:
# https://github.com/PacificBiosciences/pbbioconda/wiki/IPA-Documentation
import os
print(f'CWD:{os.getcwd()}')

CWD = os.getcwd()
CTG_PREFIX = "ctg."

cfg = config
READS_FN = cfg['reads_fn']
GENOME_SIZE = cfg['genome_size']
COVERAGE = cfg['coverage']
ADVANCED_OPTIONS = cfg["advanced_options"]
POLISH_RUN = 1 if cfg['polish_run'] else 0
PHASING_RUN = 1 if cfg['phase_run'] else 0
PURGE_DUPS_RUN = cfg.get('purge_dups_run', 0) # temporary
PURGE_DUPS_CALCUTS = cfg.get('purge_dups_calcuts', "") # temporary
M4FILT_HIGH_COPY_SAMPLE_RATE = cfg.get('m4filt_high_copy_sample_rate', 1.0) # temporary
NPROC = cfg['nproc']
MAX_NCHUNKS = 40 if 'max_nchunks' not in cfg else cfg['max_nchunks']
TMP_DIR = '/tmp' if 'tmp_dir' not in cfg else cfg['tmp_dir']

LOG_LEVEL = "INFO"
READS_DB_PREFIX = "reads"
CONTIGS_DB_PREFIX = "contigs"

# until we find a way to use NPROC*workflow.cores for a local-only serial task
NPROC_SERIAL = NPROC

print(f'NPROC:{NPROC}')
print(f'NPROC_SERIAL:{NPROC_SERIAL}')

QSUB_LOG = 'qsub_log' # directory
if not os.path.isdir(QSUB_LOG):
    os.makedirs(QSUB_LOG)

if not os.path.isabs(READS_FN):
    READS_FN = os.path.abspath(os.path.join(CWD, '..', READS_FN))

print(f'config:{config}')
if os.environ.get('IPA_QUIET'):
    shell.prefix("set -eu -o pipefail; ")
else:
    shell.prefix("set -vxeu -o pipefail; ")


# Choose are directory names with sequential prefixes.
# (Assume fewer than 100, LOL.)

RULE_COUNT = 0

def prenum():
    global RULE_COUNT
    RULE_COUNT += 1
    assert RULE_COUNT < 100
    return f'{RULE_COUNT:02d}-'

generate_config_dn = prenum() + 'generate_config'
build_db_dn = prenum() + 'build_db'
ovl_asym_prepare_dn = prenum() + 'ovl_asym_prepare'
ovl_asym_run_dn = prenum() + 'ovl_asym_run'
ovl_asym_merge_dn = prenum() + 'ovl_asym_merge'
phasing_prepare_dn = prenum() + 'phasing_prepare'
phasing_run_dn = prenum() + 'phasing_run'
phasing_merge_dn = prenum() + 'phasing_merge'
ovl_filter_dn = prenum() + 'ovl_filter'
assemble_dn = prenum() + 'assemble'
polish_prepare_dn = prenum() + 'polish_prepare'
polish_run_dn = prenum() + 'polish_run'
polish_merge_dn = prenum() + 'polish_merge'
separate_dn = prenum() + 'separate'
purge_dups_map_prepare_dn = prenum() + 'purge_dups_map_prepare'
purge_dups_map_run_dn = prenum() + 'purge_dups_map_run'
purge_dups_map_merge_dn = prenum() + 'purge_dups_map_merge'
purge_dups_paf_dn = prenum() + 'purge_dups'
final_dn = prenum() + 'final'

localrules: generate_config, ovl_asym_prepare, separate_p_from_a, final, finish

join = os.path.join

P_CTG_FASTA = join(final_dn, "final.p_ctg.fasta")
A_CTG_FASTA = join(final_dn, "final.a_ctg.fasta")

rule finish:
    input:
        P_CTG_FASTA,
        A_CTG_FASTA,

rule generate_config:
    output:
        config = join(generate_config_dn, "generated.config"),
    input:
        reads_fn = READS_FN,
    params:
        num_threads = 1, # not needed for localrule, but does not hurt
        genome_size = GENOME_SIZE,
        coverage = COVERAGE,
        advanced_opt_str = ADVANCED_OPTIONS,
        polish_run = POLISH_RUN,
        phase_run = PHASING_RUN,
        purge_dups_run = PURGE_DUPS_RUN,
        purge_dups_calcuts = PURGE_DUPS_CALCUTS,
        m4filt_high_copy_sample_rate = M4FILT_HIGH_COPY_SAMPLE_RATE,
        log_level = LOG_LEVEL,
        tmp_dir = TMP_DIR,
    shell: """
        wd=$(dirname {output.config})
        mkdir -p $wd
        cd $wd
        rel=..

        params_advanced_opt="{params.advanced_opt_str}" \
        params_coverage="{params.coverage}" \
        params_genome_size="{params.genome_size}" \
        params_polish_run="{params.polish_run}" \
        params_phase_run="{params.phase_run}" \
        params_purge_dups_run="{params.purge_dups_run}" \
        params_purge_dups_calcuts="{params.purge_dups_calcuts}" \
        params_m4filt_high_copy_sample_rate="{params.m4filt_high_copy_sample_rate}" \
        params_log_level="{params.log_level}" \
        params_tmp_dir="{params.tmp_dir}" \
        output_fn="generated.config.sh" \
        sentinel_fn="generated.config" \
            time ipa2-task generate_config_from_workflow
    """

rule build_db:
    output:
        seqdb = join(build_db_dn, "reads.seqdb"),
        seeddb = join(build_db_dn, "reads.seeddb"),
        seqdb_seqs = join(build_db_dn, "reads.seqdb.0.seq"),
        seeddb_seeds = join(build_db_dn, "reads.seeddb.0.seeds"),
        input_fofn = join(build_db_dn, "input.fofn"),
    input:
        reads_fn = READS_FN,
        config_sh_fn = rules.generate_config.output.config,
    params:
        num_threads = NPROC_SERIAL,
        db_prefix = READS_DB_PREFIX,
        log_level = LOG_LEVEL,
        tmp_dir = TMP_DIR,
    shell: """
        wd=$(dirname {output.seqdb})
        mkdir -p $wd
        cd $wd
        rel=..

        input_reads_fn="{input.reads_fn}" \
        params_db_prefix="{params.db_prefix}" \
        params_config_sh_fn="$rel/{input.config_sh_fn}" \
        params_num_threads={params.num_threads} \
        params_log_level="{params.log_level}" \
        params_tmp_dir="{params.tmp_dir}" \
            time ipa2-task build_db
    """

checkpoint ovl_asym_prepare:
    output:
        sharddir = directory(join(ovl_asym_prepare_dn, "shards")),
    input:
        seqdb = rules.build_db.output.seqdb,
    params:
        num_threads = 1, # not needed for localrule, but does not hurt
        max_nchunks = MAX_NCHUNKS,
        log_level = LOG_LEVEL,
        tmp_dir = TMP_DIR,
    shell: """
        rm -f {output.sharddir}
        mkdir -p {output.sharddir}
        cd {output.sharddir}
        rel=../..

        input_db="$rel/{input.seqdb}" \
        output_shard_ids=./all_shard_ids \
        output_pwd=./pwd \
        params_max_nchunks="{params.max_nchunks}" \
        params_log_level="{params.log_level}" \
        params_tmp_dir="{params.tmp_dir}" \
            time ipa2-task ovl_asym_prepare
    """

def gathered_m4(wildcards):
    checkpoint_output = checkpoints.ovl_asym_prepare.get(**wildcards).output.sharddir  # raises until checkpoint is done
    shard_ids_ovl = glob_wildcards(os.path.join(checkpoint_output, "shard.{shard_id_ovl}.pancake_ranges")).shard_id_ovl
    return expand(join(ovl_asym_run_dn, "{shard_id_ovl}/ovl.sorted.m4"),
            shard_id_ovl=shard_ids_ovl)

rule ovl_asym_run:
    output:
        out_m4 = temp(join(ovl_asym_run_dn, "{shard_id_ovl}/ovl.m4")),
        out_sorted_m4 = temp(join(ovl_asym_run_dn, "{shard_id_ovl}/ovl.sorted.m4")),
    input:
        seqdb = rules.build_db.output.seqdb,
        seeddb = rules.build_db.output.seeddb,
        seqdb_seqs = rules.build_db.output.seqdb_seqs,
        seeddb_seeds = rules.build_db.output.seeddb_seeds,
        config_sh_fn = rules.generate_config.output.config,
    params:
        num_threads = NPROC,
        log_level = LOG_LEVEL,
        tmp_dir = TMP_DIR,
        db_prefix = READS_DB_PREFIX,
        ovl_asym_prepare_dn = rules.ovl_asym_prepare.output.sharddir,
    shell: """
        echo "shard_id={wildcards.shard_id_ovl}"
        wd=$(dirname {output.out_m4})
        mkdir -p $wd
        cd $wd
        rel=../..

        input_seqdb="$rel/{input.seqdb}" \
        params_ovl_asym_prepare_dn="$rel/{params.ovl_asym_prepare_dn}" \
        params_shard_id={wildcards.shard_id_ovl} \
        params_db_prefix="{params.db_prefix}" \
        params_num_threads={params.num_threads} \
        params_config_sh_fn="$rel/{input.config_sh_fn}" \
        params_log_level="{params.log_level}" \
        params_tmp_dir="{params.tmp_dir}" \
            time ipa2-task ovl_asym_run
    """

rule ovl_asym_merge:
    output:
        m4_merged_raw = join(ovl_asym_merge_dn, "ovl.merged.m4"),
        m4_filtered_nonlocal = join(ovl_asym_merge_dn, "ovl.nonlocal.m4"),
    input:
        in_fns = gathered_m4,
        config_sh_fn = rules.generate_config.output.config,
    params:
        num_threads = NPROC_SERIAL,
        db_prefix = READS_DB_PREFIX,
        log_level = LOG_LEVEL,
        tmp_dir = TMP_DIR,
    shell: """
        wd=$(dirname {output.m4_merged_raw})
        mkdir -p $wd
        cd $wd
        rel=..

        # We must change rel-path names when we chdir.
        for fn in {input.in_fns}; do
            echo $rel/$fn
        done >| ./sorted.fofn

        input_fofn=./sorted.fofn \
        params_num_threads="{params.num_threads}" \
        params_db_prefix={params.db_prefix} \
        params_config_sh_fn="$rel/{input.config_sh_fn}" \
        params_log_level="{params.log_level}" \
        params_tmp_dir="{params.tmp_dir}" \
            time ipa2-task ovl_asym_merge
    """

#################
### Phasing #####
#################
checkpoint phasing_prepare:
    output:
        sharddir = directory(join(phasing_prepare_dn, "shards")),
    input:
        seqdb = rules.build_db.output.seqdb,
        m4 = rules.ovl_asym_merge.output.m4_filtered_nonlocal,
        config_sh_fn = rules.generate_config.output.config,
    params:
        num_threads = 1,
        max_nchunks = MAX_NCHUNKS,
        log_level = LOG_LEVEL,
        tmp_dir = TMP_DIR,
    shell: """
        rm -f {output.sharddir}
        mkdir -p {output.sharddir}
        cd {output.sharddir}
        rel=../..

        input_m4="$rel/{input.m4}" \
        output_shard_ids=./all_shard_ids \
        output_pwd=./pwd.txt \
        params_config_sh_fn="$rel/{input.config_sh_fn}" \
        params_max_nchunks="{params.max_nchunks}" \
        params_log_level="{params.log_level}" \
        params_tmp_dir="{params.tmp_dir}" \
            time ipa2-task phasing_prepare
    """

def gathered_prepared_phasing_m4(wildcards):
    checkpoint_output = checkpoints.phasing_prepare.get(**wildcards).output.sharddir  # raises until checkpoint is done
    shard_ids_phasing = glob_wildcards(os.path.join(checkpoint_output, "shard.{shard_id_phasing}.block_ids")).shard_id_phasing
    return expand(join(phasing_run_dn, "{shard_id_phasing}/outdir_fn.txt"),
            shard_id_phasing=shard_ids_phasing)

rule phasing_run:
    output:
        out_keep_m4 = join(phasing_run_dn, "{shard_id_phasing}/ovl.phased.m4"),
        out_scraps_m4 = join(phasing_run_dn, "{shard_id_phasing}/ovl.phased.m4.scraps"),
        out_outdir_fn = join(phasing_run_dn, "{shard_id_phasing}/outdir_fn.txt"),
    input:
        config_sh_fn = rules.generate_config.output.config,
        seqdb = rules.build_db.output.seqdb,
        seqdb_seqs = rules.build_db.output.seqdb_seqs,
    params:
        num_threads = NPROC,
        phasing_prepare_dn = rules.phasing_prepare.output.sharddir,
        log_level = LOG_LEVEL,
        tmp_dir = TMP_DIR,
    shell: """
        echo "shard_id={wildcards.shard_id_phasing}"
        wd=$(dirname {output.out_keep_m4})
        mkdir -p $wd
        cd $wd
        rel=../..

        input_seqdb="$rel/{input.seqdb}" \
        output_keep_m4=$(basename {output.out_keep_m4}) \
        output_scraps_m4=$(basename {output.out_scraps_m4}) \
        output_outdir_fn="outdir_fn.txt" \
        params_phasing_prepare_dn="$rel/{params.phasing_prepare_dn}" \
        params_shard_id={wildcards.shard_id_phasing} \
        params_num_threads={params.num_threads} \
        params_config_sh_fn="$rel/{input.config_sh_fn}" \
        params_log_level="{params.log_level}" \
        params_tmp_dir="{params.tmp_dir}" \
            time ipa2-task phasing_run

    """

rule phasing_merge:
    output:
        gathered_m4 = join(phasing_merge_dn, "ovl.phased.m4"),
    input:
        original_m4 = rules.ovl_asym_merge.output.m4_filtered_nonlocal,
        fns = gathered_prepared_phasing_m4,
        config_sh_fn = rules.generate_config.output.config,
    params:
        num_threads = NPROC_SERIAL,
        log_level = LOG_LEVEL,
        tmp_dir = TMP_DIR,
    shell: """
        wd=$(dirname {output.gathered_m4})
        mkdir -p $wd
        cd $wd
        rel=..

        # Collect all overlap paths to keep.
        for fn in {input.fns}; do
            # Get the folder name.
            dn=$(cat $rel/$fn)
            echo $dn/ovl.phased.m4
        done >| ./merged.keep.fofn

        # Collect all scraps paths.
        for fn in {input.fns}; do
            # Get the folder name.
            dn=$(cat $rel/$fn)
            echo $dn/ovl.phased.m4.scraps
        done >| ./merged.scraps.fofn

        input_keep_fofn="./merged.keep.fofn" \
        input_scraps_fofn="./merged.scraps.fofn" \
        input_original_m4="$rel/{input.original_m4}" \
        output_m4="ovl.phased.m4" \
        params_num_threads={params.num_threads} \
        params_config_sh_fn="$rel/{input.config_sh_fn}" \
        params_log_level="{params.log_level}" \
        params_tmp_dir="{params.tmp_dir}" \
            time ipa2-task phasing_merge
    """

rule ovl_filter:
    output:
        m4_final = join(ovl_filter_dn, "ovl.final.m4"),
        m4_chimerfilt = join(ovl_filter_dn, "ovl.chimerfilt.m4"),
    input:
        m4 = rules.phasing_merge.output.gathered_m4,
        config_sh_fn = rules.generate_config.output.config,
    params:
        num_threads = NPROC_SERIAL,
        log_level = LOG_LEVEL,
        tmp_dir = TMP_DIR,
    shell: """
        wd=$(dirname {output.m4_final})
        mkdir -p $wd
        cd $wd
        rel=..

        input_m4="$rel/{input.m4}" \
        output_m4_final=$(basename {output.m4_final}) \
        output_m4_chimerfilt=$(basename {output.m4_chimerfilt}) \
        params_num_threads={params.num_threads} \
        params_config_sh_fn="$rel/{input.config_sh_fn}" \
        params_log_level="{params.log_level}" \
        params_tmp_dir="{params.tmp_dir}" \
            time ipa2-task ovl_filter
    """

rule assemble:
    output:
        p_ctg_fasta = join(assemble_dn, "p_ctg.fasta"),
        a_ctg_fasta = join(assemble_dn, "a_ctg.fasta"),
        p_ctg_tiling_path = join(assemble_dn, "p_ctg_tiling_path"),
        a_ctg_tiling_path = join(assemble_dn, "a_ctg_tiling_path"),
        p_ctg_fa_fai = join(assemble_dn, "p_ctg.fasta.fai"),
        a_ctg_fa_fai = join(assemble_dn, "a_ctg.fasta.fai"),
        read_to_contig = join(assemble_dn, "read_to_contig.csv"),
#        asm_gfa = join(assemble_dn, "asm.gfa"),
#        sg_gfa = join(assemble_dn, "sg.gfa"),
#        contig_gfa2 = join(assemble_dn, "contig.gfa2"),
        circular_contigs = join(assemble_dn, "circular_contigs.csv"),
    input:
        reads_fn = READS_FN,
        seqdb = rules.build_db.output.seqdb,
        m4 = rules.ovl_filter.output.m4_final,
        m4_phasing_merge = rules.phasing_merge.output.gathered_m4,  # Needed for read tracking.
        config_sh_fn = rules.generate_config.output.config,
    params:
        num_threads = NPROC_SERIAL,
        ctg_prefix = CTG_PREFIX,
        log_level = LOG_LEVEL,
        tmp_dir = TMP_DIR,
    shell: """
        wd=$(dirname {output.p_ctg_fasta})
        mkdir -p $wd
        cd $wd
        rel=..

        input_seqdb="$rel/{input.seqdb}" \
        input_m4="$rel/{input.m4}" \
        input_m4_phasing_merge="$rel/{input.m4_phasing_merge}" \
        input_reads="{input.reads_fn}" \
        params_ctg_prefix={params.ctg_prefix} \
        params_num_threads={params.num_threads} \
        params_config_sh_fn="$rel/{input.config_sh_fn}" \
        params_log_level="{params.log_level}" \
        params_tmp_dir="{params.tmp_dir}" \
            time ipa2-task assemble
    """

#################
### Polishing ###
#################
checkpoint polish_prepare:
    output:
        sharddir = directory(join(polish_prepare_dn, "shards")),
    input:
        read_to_contig = rules.assemble.output.read_to_contig,
        config_sh_fn = rules.generate_config.output.config,
        # The p_ctg and a_ctg FASTA files are needed to enable the
        # config 'polish' option to skip polishing. If polishing is skipped
        # then the output "polished" contigs will be just the draft sequences.
        p_ctg_fasta_fai = rules.assemble.output.p_ctg_fa_fai,
        a_ctg_fasta_fai = rules.assemble.output.a_ctg_fa_fai,
    params:
        num_threads = 1, # not needed for localrule, but does not hurt
        max_nchunks = MAX_NCHUNKS,
        log_level = LOG_LEVEL,
        tmp_dir = TMP_DIR,
    shell: """
        rm -f {output.sharddir}
        mkdir -p {output.sharddir}
        cd {output.sharddir}
        rel=../..

        input_read_to_contig="$rel/{input.read_to_contig}" \
        input_p_ctg_fasta_fai="$rel/{input.p_ctg_fasta_fai}" \
        input_a_ctg_fasta_fai="$rel/{input.a_ctg_fasta_fai}" \
        output_shard_ids=./all_shard_ids \
        output_pwd=./pwd.txt \
        params_config_sh_fn="$rel/{input.config_sh_fn}" \
        params_max_nchunks="{params.max_nchunks}" \
        params_log_level="{params.log_level}" \
        params_tmp_dir="{params.tmp_dir}" \
            time ipa2-task polish_prepare
    """
def gathered_polish(wildcards):
    checkpoint_output = checkpoints.polish_prepare.get(**wildcards).output.sharddir  # raises until checkpoint is done
    shard_ids_polish = glob_wildcards(os.path.join(checkpoint_output, "shard.{shard_id_polish}.block_ids")).shard_id_polish
    return expand(join(polish_run_dn, "{shard_id_polish}/consensus.fasta"),
            shard_id_polish=shard_ids_polish)

rule polish_run:
    output:
        consensus = temp(join(polish_run_dn, "{shard_id_polish}/consensus.fasta")),
    input:
        input_fofn = rules.build_db.output.input_fofn,
        seqdb = rules.build_db.output.seqdb,
        seqdb_seqs = rules.build_db.output.seqdb_seqs,
        p_ctg_fasta = rules.assemble.output.p_ctg_fasta,
        a_ctg_fasta = rules.assemble.output.a_ctg_fasta,
        config_sh = rules.generate_config.output.config, # could be a param if we want to regen
    params:
        num_threads = NPROC,
        polish_prepare_dn = rules.polish_prepare.output.sharddir,
        log_level = LOG_LEVEL,
        tmp_dir = TMP_DIR,
    shell: """
        echo "shard_id={wildcards.shard_id_polish}"
        wd=$(dirname {output.consensus})
        mkdir -p $wd
        cd $wd
        rel=../..

        input_fofn="$rel/{input.input_fofn}" \
        input_seqdb="$rel/{input.seqdb}" \
        input_p_ctg_fasta="$rel/{input.p_ctg_fasta}" \
        input_a_ctg_fasta="$rel/{input.a_ctg_fasta}" \
        output_consensus=$(basename {output.consensus}) \
        params_polish_prepare_dn="$rel/{params.polish_prepare_dn}" \
        params_shard_id={wildcards.shard_id_polish} \
        params_num_threads={params.num_threads} \
        params_config_sh_fn="$rel/{input.config_sh}" \
        params_log_level="{params.log_level}" \
        params_tmp_dir="{params.tmp_dir}" \
            time ipa2-task polish_run
    """

rule polish_merge:
    output:
        consensus_merged = join(polish_merge_dn, "assembly.merged.fasta"),
        consensus_merged_fai = join(polish_merge_dn, "assembly.merged.fasta.fai"),
    input:
        fns = gathered_polish,
        p_ctg_fasta = rules.assemble.output.p_ctg_fasta,
        a_ctg_fasta = rules.assemble.output.a_ctg_fasta,
        config_sh_fn = rules.generate_config.output.config,
    params:
        num_threads = NPROC_SERIAL,
        log_level = LOG_LEVEL,
        tmp_dir = TMP_DIR,
    shell: """
        wd=$(dirname {output.consensus_merged})
        mkdir -p $wd
        cd $wd
        rel=..

        # We must change rel-path names when we chdir.
        for fn in {input.fns}; do
            echo $rel/$fn
        done >| ./merged.fofn

        for fn in {input.fns}; do
            dn=$(dirname $rel/$fn)
            echo $dn
        done >| ./merged.fodn

        input_fofn=./merged.fofn \
        input_fodn=./merged.fodn \
        input_p_ctg_fasta="$rel/{input.p_ctg_fasta}" \
        input_a_ctg_fasta="$rel/{input.a_ctg_fasta}" \
        output_fasta=$(basename {output.consensus_merged}) \
        output_p_paf="p.paf" \
        output_a_paf="a.paf" \
        params_num_threads="{params.num_threads}" \
        params_config_sh_fn="$rel/{input.config_sh_fn}" \
        params_log_level="{params.log_level}" \
        params_tmp_dir="{params.tmp_dir}" \
            time ipa2-task polish_merge
    """

rule separate_p_from_a:
    output:
        p_ctg_fasta = join(separate_dn, "p_ctg.fasta"),
        a_ctg_fasta = join(separate_dn, "a_ctg.fasta"),
    input:
        assembly_merged_fasta = rules.polish_merge.output.consensus_merged,
        assembly_merged_fai = rules.polish_merge.output.consensus_merged_fai,
        p_ctg_fasta = rules.assemble.output.p_ctg_fasta,
        a_ctg_fasta = rules.assemble.output.a_ctg_fasta,
    params:
        num_threads = 1, # not needed for localrule, but does not hurt
        log_level = LOG_LEVEL,
        tmp_dir = TMP_DIR,
        polish_run = POLISH_RUN, 
    shell: """
        wd=$(dirname {output.p_ctg_fasta})
        mkdir -p $wd
        cd $wd
        rel=..

        input_assembly_merged_fasta="$rel/{input.assembly_merged_fasta}" \
        input_p_ctg_fasta="$rel/{input.p_ctg_fasta}" \
        input_a_ctg_fasta="$rel/{input.a_ctg_fasta}" \
        output_p_ctg_fasta=$(basename {output.p_ctg_fasta}) \
        output_a_ctg_fasta=$(basename {output.a_ctg_fasta}) \
        params_log_level="{params.log_level}" \
        params_tmp_dir="{params.tmp_dir}" \
        params_polish_run="{params.polish_run}" \
            time ipa2-task separate_p_from_a
    """

checkpoint purge_dups_map_prepare:
    output:
        sharddir = directory(join(purge_dups_map_prepare_dn, "shards")),
    input:
        seqdb = rules.build_db.output.seqdb,
        p_ctg_fasta = rules.separate_p_from_a.output.p_ctg_fasta,
        config_sh_fn = rules.generate_config.output.config,
    params:
        num_threads = NPROC,
        max_nchunks = MAX_NCHUNKS,
        log_level = LOG_LEVEL,
    shell: """
        rm -f {output.sharddir}
        mkdir -p {output.sharddir}
        cd {output.sharddir}
        rel=../..

        input_reads_db="$rel/{input.seqdb}" \
        input_primary_fasta="$rel/{input.p_ctg_fasta}" \
        output_shard_ids=./all_shard_ids \
        output_pwd=./pwd \
        params_num_threads="{params.num_threads}" \
        params_max_nchunks="{params.max_nchunks}" \
        params_log_level="{params.log_level}" \
        params_config_sh_fn="$rel/{input.config_sh_fn}" \
            time ipa2-task purge_dups_map_prepare
    """

def gathered_paf(wildcards):
    checkpoint_output = checkpoints.purge_dups_map_prepare.get(**wildcards).output.sharddir  # raises until checkpoint is done
    shard_ids_purge_dups_map = glob_wildcards(os.path.join(checkpoint_output, "shard.{shard_id_purge_dups_map}.pancake_ranges")).shard_id_purge_dups_map
    return expand(join(purge_dups_map_run_dn, "{shard_id_purge_dups_map}/ovl.paf.gz"),
            shard_id_purge_dups_map=shard_ids_purge_dups_map)

rule purge_dups_map_run:
    output:
        out_paf = join(purge_dups_map_run_dn, "{shard_id_purge_dups_map}/ovl.paf.gz"),
    input:
        seqdb = rules.build_db.output.seqdb,
        config_sh_fn = rules.generate_config.output.config,
    params:
        num_threads = NPROC,
        log_level = LOG_LEVEL,
        purge_dups_map_prepare_dn = rules.purge_dups_map_prepare.output.sharddir,
    shell: """
        echo "shard_id={wildcards.shard_id_purge_dups_map}"
        wd=$(dirname {output.out_paf})
        mkdir -p $wd
        cd $wd
        rel=../..

        input_seqdb="$rel/{input.seqdb}" \
        params_purge_dups_map_prepare_dn="$rel/{params.purge_dups_map_prepare_dn}" \
        params_shard_id={wildcards.shard_id_purge_dups_map} \
        params_num_threads={params.num_threads} \
        params_config_sh_fn="$rel/{input.config_sh_fn}" \
        params_log_level="{params.log_level}" \
        output_paf="$(basename {output.out_paf})" \
            time ipa2-task purge_dups_map_run
    """

rule purge_dups_map_merge:
    output:
        merged_paf = join(purge_dups_map_merge_dn, "merged.paf.gz"),
    input:
        in_fns = gathered_paf,
        config_sh_fn = rules.generate_config.output.config,
    params:
        num_threads = NPROC,
        log_level = LOG_LEVEL,
    shell: """
        wd=$(dirname {output.merged_paf})
        mkdir -p $wd
        cd $wd
        rel=..

        # We must change rel-path names when we chdir.
        for fn in {input.in_fns}; do
            echo $rel/$fn
        done >| ./sorted.fofn

        input_fofn=./sorted.fofn \
        params_num_threads="{params.num_threads}" \
        params_config_sh_fn="$rel/{input.config_sh_fn}" \
        params_log_level="{params.log_level}" \
        output_paf="$(basename {output.merged_paf})" \
            time ipa2-task purge_dups_map_merge
    """

rule purge_dups_paf:
    output:
        p_ctg_fasta = join(purge_dups_paf_dn, "final_purged_primary.fasta"),
        a_ctg_fasta = join(purge_dups_paf_dn, "final_purged_haplotigs.fasta"),
    input:
        paf = rules.purge_dups_map_merge.output.merged_paf,
        p_ctg_fasta = rules.separate_p_from_a.output.p_ctg_fasta,
        a_ctg_fasta = rules.separate_p_from_a.output.a_ctg_fasta,
        config_sh_fn = rules.generate_config.output.config,
    params:
        num_threads = NPROC,
        log_level = LOG_LEVEL,
    shell: """
        wd=$(dirname {output.p_ctg_fasta})
        mkdir -p $wd
        cd $wd
        rel=..

        input_paf="$rel/{input.paf}" \
        input_primary_fasta="$rel/{input.p_ctg_fasta}" \
        input_haplotigs_fasta="$rel/{input.a_ctg_fasta}" \
        params_num_threads="{params.num_threads}" \
        params_log_level="{params.log_level}" \
        params_config_sh_fn="$rel/{input.config_sh_fn}" \
        output_primary_fasta=$(basename {output.p_ctg_fasta}) \
        output_haplotigs_fasta=$(basename {output.a_ctg_fasta}) \
            time ipa2-task purge_dups_paf
    """

# For simplicity, we copy the minimal results into their own rule-directory.
rule final:
    output:
        p_ctg_fasta = P_CTG_FASTA,
        a_ctg_fasta = A_CTG_FASTA,
    input:
        p_ctg_fasta = rules.purge_dups_paf.output.p_ctg_fasta,
        a_ctg_fasta = rules.purge_dups_paf.output.a_ctg_fasta,
    shell: """
        wd=$(dirname {output.p_ctg_fasta})
        mkdir -p $wd
        cd $wd
        rel=..

        input_p_ctg_fasta="$rel/{input.p_ctg_fasta}" \
        input_a_ctg_fasta="$rel/{input.a_ctg_fasta}" \
        output_p_ctg_fasta="$rel/{output.p_ctg_fasta}" \
        output_a_ctg_fasta="$rel/{output.a_ctg_fasta}" \
            time ipa2-task final_copy
    """
